'''
Provides a class for interacting with the GPIO pins through I2C and GPIO peripheral
'''


import logging
from edgepi.gpio.gpio_configs import generate_expander_pin_info
from edgepi.gpio.gpio_constants import GpioDevPaths
from edgepi.peripherals.i2c import I2CDevice
from edgepi.reg_helper.reg_helper import OpCode, apply_opcodes, is_bit_set

_logger = logging.getLogger(__name__)

# pylint: disable=logging-too-many-args
class EdgePiGPIOExpander(I2CDevice):
    '''
    A class used to represent the GPIO Expander configuration for an I2C Device.
    This class will be imported to each module that requires GPIO manipulation.
    It is not intended for users.
    '''
    def __init__(self):
        super().__init__(GpioDevPaths.I2C_DEV_PATH.value)
        # get this device's expander pin names and opcodes for set, clear, direction ops
        self.expander_pin_dict = generate_expander_pin_info()

    def __read_register(self, reg_address: int, dev_address: int) -> int:
        '''
        function to read one register value from an I2C device
        Args:
            `reg_address`: register/port address to read data from
            `dev_address`: expander address to read
        Returns:
            `int`: 8-bit uint value of port/register
        '''
        msg_read = self.set_read_msg(reg_address, [0xFF])
        _logger.debug(f'Read Message: Register Address {msg_read[0].data}\
                      , Msg Place Holder {msg_read[1].data}')
        self.transfer(dev_address, msg_read)
        _logger.debug(f'Message Read: Register Address {msg_read[0].data},\
         Msg Place Holder {msg_read[1].data}')
        return msg_read[1].data[0]

    def __write_changed_values(self, reg_dict: dict, dev_address: int):
        '''
        Function to write changed values to the specified register
        In:
            reg_dict (dict): register address to value and is_changed flag
                             {register_address : {'value' : value(int), is_changed : bool}}
            dev_address: device address, 32 or 33
        Returns:
            void
        '''
        for reg_addx, entry in reg_dict.items():
            if entry['is_changed']:
                msg_write = self.set_write_msg(reg_addx, [entry['value']])
                _logger.debug(f'Write Message Content {msg_write[0]}')
                self.transfer(dev_address, msg_write)

    def read_expander_pin(self, pin_name: str) -> bool:
        '''
        Get the current state of a GPIO expander pin (low or high).

        Args:
            `pin_name` (str): name of the pin whose state to read

        Returns:
            `bool`: True if state is high, False if state is low
        '''
        dev_address = self.expander_pin_dict[pin_name].address
        reg_addx = self.expander_pin_dict[pin_name].set_code.reg_address
        pin_mask = self.expander_pin_dict[pin_name].set_code.op_mask

        # read register at reg_addx
        reg_val = self.__read_register(reg_addx, dev_address)
        pin_state = is_bit_set(reg_val, pin_mask)
        _logger.debug(":read_expander_pin: pin '%s' = '%s'", pin_name,  pin_state)

        return pin_state

    def get_expander_pin_direction(self, pin_name: str) -> bool:
        '''
        Get the current direction of a GPIO expander pin (low or high).

        Args:
            `pin_name` (str): name of the pin whose state to read

        Returns:
            `bool`: True if direction is input, False if direction is output
        '''
        dev_address = self.expander_pin_dict[pin_name].address
        # dir_out_code and dir_in_code have the same addx and mask, doesn't matter which
        reg_addx = self.expander_pin_dict[pin_name].dir_out_code.reg_address
        pin_mask = self.expander_pin_dict[pin_name].dir_out_code.op_mask

        # read register at reg_addx
        reg_val = self.__read_register(reg_addx, dev_address)
        pin_state = is_bit_set(reg_val, pin_mask)
        _logger.debug(":get_expander_pin_direction: pin '%s' = '%s'", pin_name,  pin_state)

        return pin_state

    def set_expander_pin_direction_out(self, pin_name: str):
        '''
        Set the direction of a GPIO expander pin to output. Note this will
        set pin to low before setting to output for safety reasons.

        Args:
            `pin_name` (str): name of the pin whose direction to set
        '''
        dev_address = self.expander_pin_dict[pin_name].address
        dir_out_code = self.expander_pin_dict[pin_name].dir_out_code
        reg_addx = dir_out_code.reg_address

        # get register value of port this pin belongs to
        reg_val = self.__read_register(reg_addx, dev_address)

        # set pin to low before setting to output (hazard)
        self.clear_expander_pin(pin_name)

        # set pin direction to out
        self.__apply_code_to_register(dev_address, reg_addx, reg_val, dir_out_code)
        _logger.debug(":set_expander_pin_direction_out: pin '%s' set to output", pin_name)

        self.expander_pin_dict[pin_name].is_out = True

    def set_expander_pin_direction_in(self, pin_name: str):
        '''
        Set the direction of a GPIO expander pin to high impedance input.

        Args:
            `pin_name` (str): name of the pin whose direction to set
        '''
        dev_address = self.expander_pin_dict[pin_name].address
        dir_in_code = self.expander_pin_dict[pin_name].dir_in_code
        reg_addx = dir_in_code.reg_address

        # get register value of port this pin belongs to
        reg_val = self.__read_register(reg_addx, dev_address)

        # set pin direction to input
        self.__apply_code_to_register(dev_address, reg_addx, reg_val, dir_in_code)
        _logger.debug(":set_expander_pin_direction_in: pin '%s' set to output", pin_name)

        self.expander_pin_dict[pin_name].is_out = False

    def set_expander_pin(self, pin_name: str):
        '''
        Function set gpio pin state to high

        Args:
            `pin_name` (str): name of the pin to set
        '''
        dev_address = self.expander_pin_dict[pin_name].address
        set_code = self.expander_pin_dict[pin_name].set_code
        reg_addx = set_code.reg_address

        # get register value of port this pin belongs to
        reg_val = self.__read_register(reg_addx, dev_address)

        # set pin direction to output (also sets to low)
        self.set_expander_pin_direction_out(pin_name)

        # set pin state to high
        self.__apply_code_to_register(dev_address, reg_addx, reg_val, set_code)
        _logger.debug(":set_expander_pin: pin '%s' = set to high", pin_name)

        self.expander_pin_dict[pin_name].is_high = True

    def __apply_code_to_register(self, dev_addx: int, reg_addx: int, reg_val: int, opcode: OpCode):
        """
        Applies an opcode obtained from I2CPinInfo object to a register.

        Args:
            `dev_addx` (int): I2C device address
            `reg_addx` (int): register/port address
            `reg_val` (int): current 8-bit value of register at reg_addx
            `opcode` (OpCode): opcode to apply to register at reg_addx
        """
        # apply opcode to get new register value
        reg_map = {reg_addx: reg_val}
        updated_reg_map = apply_opcodes(reg_map, [opcode])
        # write new register value to set pin new state
        self.__write_changed_values(updated_reg_map, dev_addx)
        _logger.debug(
            "__apply_code_to_register: dev_addx='%s', reg_addx='%s', reg_val='%s, opcode='%s'",
            dev_addx,
            hex(reg_addx),
            hex(reg_val),
            opcode
        )

    def clear_expander_pin(self, pin_name: str):
        '''
        Function clear gpio pin state to low
        In:
            `pin_name` (str): name of the pin to set
        '''
        dev_address = self.expander_pin_dict[pin_name].address
        clear_code = self.expander_pin_dict[pin_name].clear_code
        reg_addx = clear_code.reg_address

        # get register value of port this pin belongs to
        reg_val = self.__read_register(reg_addx, dev_address)

        # set pin state to low
        self.__apply_code_to_register(dev_address, reg_addx, reg_val, clear_code)
        _logger.debug(":set_expander_pin: pin '%s' = set to low", pin_name)

        self.expander_pin_dict[pin_name].is_high = False

    def toggle_expander_pin(self, pin_name: str):
        '''
        Function toggle gpio pin state to opposite logic

            High -> Low,
            Low -> High

        Args:
            `pin_name` (str): name of the pin to toggle
        '''
        dev_address = self.expander_pin_dict[pin_name].address
        # set and clear codes for a pin have same reg_addx and mask
        code = self.expander_pin_dict[pin_name].clear_code
        reg_addx = code.reg_address
        pin_mask = code.op_mask

        # get register value of port this pin belongs to
        reg_val = self.__read_register(reg_addx, dev_address)

        # if pin is set to low, set to high and vice-versa
        _logger.debug(":toggle_expander_pin: toggling pin '%s'", pin_name)
        if is_bit_set(reg_val, pin_mask):
            self.clear_expander_pin(pin_name)
        else:
            self.set_expander_pin(pin_name)



'''
Module for generating calibration parameter data structure for each module. This module should be
able to generate new calibration parameters using measurements tools.
'''

from edgepi.calibration.eeprom_constants import ModuleNames
from edgepi.calibration.calibration_constants import NumOfCh


class EdgePiCalibration():
    '''
    EdgePi Calibration Class handling the following functionality
    1. load calibration parameter
    2. store new calibration parameter
    3. calibration process for each module
        - Load measurement values
           - Edge Pi measurements
           - Equipment measurements
        - Calculation: use leat mean square method
    '''

    def __init__(self, module: ModuleNames):
        self.module = module.value
        self.num_of_ch = NumOfCh[module.name].value
        self.full_scale_range = None
        self.full_scale_code = None

    def generate_measurements_dict(self, num_of_points: int = None):
        '''
        Function to generate dictionary of measurements to generate calibration parameter. It will
        store expected and actual measurements at each data point for number_of_points length
        Args:
            num_of_points (int): number of points to be generated, used as length of the dictionary
        Return:
            measurements_dict (dict): dictionary mapping expected value to
            ie) {0 : {'input_unit' : int, 'expected_out' : float, 'actual_out' : float},
                 .
                 .
                 .
                 nth_point : {'input_unit' : int, 'expected_out' : float, 'actual_out' : float}}
        '''
        measurements_dict = {}
        for point in range(num_of_points):
            measurements_dict[point]= {'input_unit': 0 ,'expected_out': 0, 'actual_out': 0}
        return measurements_dict

    def generate_channel_measurements_dict(self, num_of_points: int = None):
        '''
        Function to generate measuremenst dictionary for each channel. This is neccessary to calibr-
        -ate each channel individually
        Args:
            num_of_points (int): number of points to be generated, used as length of the dictionary
        '''
        ch_measurements_dict = {}
        for ch in range(self.num_of_ch):
            ch_measurements_dict[ch] = self.generate_measurements_dict(num_of_points)
        return ch_measurements_dict

    def record_measurements(self, nth_measurements: dict = None,
                                  input_unit: int = None,
                                  expected: float = None,
                                  actual: float = None):
        '''
        Modify the expected and actual measurements of nth item of measuremnts dictionary
        Arg:
            nth_measurements (dict): nth point in measurements dictionary
            ie) {'expected' : float, 'actual' : float}
            expected (float): expeceted measurements
            actual (float): actual measurements
        Return:
            N/A
        '''
        nth_measurements['input_unit'] = input_unit
        nth_measurements['expected_out'] = expected
        nth_measurements['actual_out'] = actual



'''unit tests for tc_faults.py module'''

import pytest
from bitstring import Bits
from edgepi.tc.tc_faults import map_fault_status, Fault, FaultMsg, FaultType


@pytest.mark.parametrize(
    "fault_bits, mask_bits, expected",
    [
        (
            Bits("0xFF"),
            Bits("0xFF"),
            {
                FaultType.CJRANGE: Fault(
                    FaultType.CJRANGE,
                    FaultMsg.CJRANGE_BAD_MSG,
                    at_fault=True,
                    is_masked=True,
                ),
                FaultType.TCRANGE: Fault(
                    FaultType.TCRANGE,
                    FaultMsg.TCRANGE_BAD_MSG,
                    at_fault=True,
                    is_masked=True,
                ),
                FaultType.CJHIGH: Fault(
                    FaultType.CJHIGH,
                    FaultMsg.CJHIGH_BAD_MSG,
                    at_fault=True,
                    is_masked=True,
                ),
                FaultType.CJLOW: Fault(
                    FaultType.CJLOW,
                    FaultMsg.CJLOW_BAD_MSG,
                    at_fault=True,
                    is_masked=True,
                ),
                FaultType.TCHIGH: Fault(
                    FaultType.TCHIGH,
                    FaultMsg.TCHIGH_BAD_MSG,
                    at_fault=True,
                    is_masked=True,
                ),
                FaultType.TCLOW: Fault(
                    FaultType.TCLOW,
                    FaultMsg.TCLOW_BAD_MSG,
                    at_fault=True,
                    is_masked=True,
                ),
                FaultType.OVUV: Fault(
                    FaultType.OVUV, FaultMsg.OVUV_BAD_MSG, at_fault=True, is_masked=True
                ),
                FaultType.OPEN: Fault(
                    FaultType.OPEN, FaultMsg.OPEN_BAD_MSG, at_fault=True, is_masked=True
                ),
            },
        ),  # all faults, all masks
        (
            Bits("0x00"),
            Bits("0xFF"),
            {
                FaultType.CJRANGE: Fault(
                    FaultType.CJRANGE, FaultMsg.CJRANGE_OK_MSG, is_masked=True
                ),
                FaultType.TCRANGE: Fault(
                    FaultType.TCRANGE, FaultMsg.TCRANGE_OK_MSG, is_masked=True
                ),
                FaultType.CJHIGH: Fault(
                    FaultType.CJHIGH, FaultMsg.CJHIGH_OK_MSG, is_masked=True
                ),
                FaultType.CJLOW: Fault(
                    FaultType.CJLOW, FaultMsg.CJLOW_OK_MSG, is_masked=True
                ),
                FaultType.TCHIGH: Fault(
                    FaultType.TCHIGH, FaultMsg.TCHIGH_OK_MSG, is_masked=True
                ),
                FaultType.TCLOW: Fault(
                    FaultType.TCLOW, FaultMsg.TCLOW_OK_MSG, is_masked=True
                ),
                FaultType.OVUV: Fault(
                    FaultType.OVUV, FaultMsg.OVUV_OK_MSG, is_masked=True
                ),
                FaultType.OPEN: Fault(
                    FaultType.OPEN, FaultMsg.OPEN_OK_MSG, is_masked=True
                ),
            },
        ),  # no faults, all masks
        (
            Bits("0x00"),
            Bits("0x00"),
            {
                FaultType.CJRANGE: Fault(
                    FaultType.CJRANGE, FaultMsg.CJRANGE_OK_MSG, is_masked=False
                ),
                FaultType.TCRANGE: Fault(
                    FaultType.TCRANGE, FaultMsg.TCRANGE_OK_MSG, is_masked=False
                ),
                FaultType.CJHIGH: Fault(
                    FaultType.CJHIGH, FaultMsg.CJHIGH_OK_MSG, is_masked=False
                ),
                FaultType.CJLOW: Fault(
                    FaultType.CJLOW, FaultMsg.CJLOW_OK_MSG, is_masked=False
                ),
                FaultType.TCHIGH: Fault(
                    FaultType.TCHIGH, FaultMsg.TCHIGH_OK_MSG, is_masked=False
                ),
                FaultType.TCLOW: Fault(
                    FaultType.TCLOW, FaultMsg.TCLOW_OK_MSG, is_masked=False
                ),
                FaultType.OVUV: Fault(
                    FaultType.OVUV, FaultMsg.OVUV_OK_MSG, is_masked=False
                ),
                FaultType.OPEN: Fault(
                    FaultType.OPEN, FaultMsg.OPEN_OK_MSG, is_masked=False
                ),
            },
        ),  # no faults, no masks
        (
            Bits("0xFF"),
            Bits("0x00"),
            {
                FaultType.CJRANGE: Fault(
                    FaultType.CJRANGE,
                    FaultMsg.CJRANGE_BAD_MSG,
                    at_fault=True,
                    is_masked=False,
                ),
                FaultType.TCRANGE: Fault(
                    FaultType.TCRANGE,
                    FaultMsg.TCRANGE_BAD_MSG,
                    at_fault=True,
                    is_masked=False,
                ),
                FaultType.CJHIGH: Fault(
                    FaultType.CJHIGH,
                    FaultMsg.CJHIGH_BAD_MSG,
                    at_fault=True,
                    is_masked=False,
                ),
                FaultType.CJLOW: Fault(
                    FaultType.CJLOW,
                    FaultMsg.CJLOW_BAD_MSG,
                    at_fault=True,
                    is_masked=False,
                ),
                FaultType.TCHIGH: Fault(
                    FaultType.TCHIGH,
                    FaultMsg.TCHIGH_BAD_MSG,
                    at_fault=True,
                    is_masked=False,
                ),
                FaultType.TCLOW: Fault(
                    FaultType.TCLOW,
                    FaultMsg.TCLOW_BAD_MSG,
                    at_fault=True,
                    is_masked=False,
                ),
                FaultType.OVUV: Fault(
                    FaultType.OVUV,
                    FaultMsg.OVUV_BAD_MSG,
                    at_fault=True,
                    is_masked=False,
                ),
                FaultType.OPEN: Fault(
                    FaultType.OPEN,
                    FaultMsg.OPEN_BAD_MSG,
                    at_fault=True,
                    is_masked=False,
                ),
            },
        ),  # all faults, no masks
    ],
)
def test_map_fault_status(fault_bits, mask_bits, expected):
    out = map_fault_status(fault_bits, mask_bits)
    assert out == expected



"""unit test for dac module"""

# pylint: disable=C0413
from unittest import mock
import sys
sys.modules['periphery'] = mock.MagicMock()

import pytest
from edgepi.dac.dac_constants import DACChannel as CH
from edgepi.dac.dac_commands import DACCommands
from edgepi.dac.dac_constants import EdgePiDacCalibrationConstants as CALIB_CONSTS, PowerMode
from edgepi.calibration.edgepi_eeprom import EdgePiEEPROM
from test_edgepi.unit_tests.test_calibration.read_serialized import read_binfile

@pytest.fixture(name="dac_ops")
def fixture_test_dac_ops(mocker):
    mocker.patch("edgepi.peripherals.i2c.I2C")
    mocker.patch("edgepi.dac.edgepi_dac.EdgePiEEPROM._EdgePiEEPROM__read_edgepi_reserved_memory",
                  return_value = read_binfile())
    eeprom = EdgePiEEPROM()
    eeprom_data  = eeprom.get_edgepi_reserved_data()
    dac_calib_params = eeprom_data.dac_calib_parms
    dac_ops = DACCommands(dac_calib_params)
    return dac_ops


# Combine command needs check for interger numbers for op-code, channel and value
# It also needs to check the range of each value.


@pytest.mark.parametrize(
    "sample, result",
    [
        ([1], True),
        ([1, 2, 3, 4, 5], True),
        ([-1, 1, 2, 3], True),
        ([444, 22, 3333, 5], True),
        ([-111, -2222], True),
    ],
)
def test_dac_check_for_int(sample, result, dac_ops):
    assert dac_ops.check_for_int(sample) == result


@pytest.mark.parametrize(
    "sample, error",
    [
        ([1, 2.22, 3, 4, 0], ValueError),
        ([None, 1, 2, 3], ValueError),
        ([], ValueError),
        ([-1, -2.22], ValueError),
    ],
)
def test_dac_check_for_int_exception(sample, error, dac_ops):
    with pytest.raises(Exception) as err:
        dac_ops.check_for_int(sample)
    assert err.type is error


@pytest.mark.parametrize(
    "range_min, target, range_max, result",
    [(0, 0, 10, True), (0, 10, 10, True), (0, 5, 10, True), (0.5, 1, 1.1, True)],
)
def test_dac_check_range(range_min, target, range_max, result, dac_ops):
    assert dac_ops.check_range(target, range_min, range_max) == result


@pytest.mark.parametrize(
    "range_min, target, range_max, error",
    [
        (0, -1, len(CH), ValueError),
        (0, 11, len(CH), ValueError),
        (0, -5, CALIB_CONSTS.RANGE.value, ValueError),
        (0, 65536, CALIB_CONSTS.RANGE.value, ValueError),
    ],
)
def test_dac_check_range_raises(range_min, target, range_max, error, dac_ops):
    with pytest.raises(Exception) as err:
        dac_ops.check_range(target, range_min, range_max)
    assert err.type is error


@pytest.mark.parametrize(
    "a, b, c, d",
    [
        (3, 1, 1000, [49, 3, 232]),
        (3, 0, 1000, [48, 3, 232]),
        (3, 3, 1000, [51, 3, 232]),
    ],
)
def test_dac_combine_command(a, b, c, d, dac_ops):
    # pylint: disable=invalid-name
    assert dac_ops.combine_command(a, b, c) == d


@pytest.mark.parametrize(
    "a, b, c",
    [(1, 1000, [49, 3, 232]), (0, 1000, [48, 3, 232]), (3, 1000, [51, 3, 232])],
)
def test_dac_generate_write_and_update_command(a, b, c, dac_ops):
    # pylint: disable=invalid-name
    assert dac_ops.generate_write_and_update_command(a, b) == c


# voltage to code conversion
# voltage = positive floating number 0~5V ideally
# code = positive integer number 0~65535
# rounding up/down during conversion ?


@pytest.mark.parametrize("ch, expected, dac_gain, result",
                        [(1, 2.345, 1, 30282), (0, 2.345, 2, 15285), (3, 2.345, 1, 30341)])
def test_dac_voltage_to_code(ch, expected, dac_gain, result, dac_ops):
    assert dac_ops.voltage_to_code(ch, expected, dac_gain) == result


@pytest.mark.parametrize(
    "ch, code, dac_gain, result",
    [
        (1, 33798, 1, 2.619),
        (0, 33798, 2, 5.239),
        (3, 33798, 1, 2.614),
    ],
)
def test_dac_code_to_voltage(ch, code, dac_gain, result, dac_ops):
    assert pytest.approx(dac_ops.code_to_voltage(ch, code, dac_gain), 0.001) == result


@pytest.mark.parametrize(
    "read_code, code_val",
    [
        ([0, 0xFF, 0xFF], 65535),
        ([0, 0, 0], 0),
        ([0, 0x75, 0x30], 30000),
    ],
)
def test_dac_extract_read_data(read_code, code_val, dac_ops):
    assert dac_ops.extract_read_data(read_code) == code_val


@pytest.mark.parametrize(
    "dac_state, expected",
    [
        ([PowerMode.NORMAL.value] * 8, 0x0000),
        ([PowerMode.POWER_DOWN_GROUND.value] + [PowerMode.NORMAL.value] * 7, 0x4000),
        ([PowerMode.POWER_DOWN_3_STATE.value] + [PowerMode.NORMAL.value] * 7, 0xC000),
        (
            [
                PowerMode.NORMAL.value,
                PowerMode.POWER_DOWN_GROUND.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
            ],
            0x1000,
        ),
        (
            [
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.POWER_DOWN_GROUND.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
            ],
            0x0400,
        ),
        (
            [
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.POWER_DOWN_GROUND.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
            ],
            0x0100,
        ),
        (
            [
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.POWER_DOWN_GROUND.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
            ],
            0x0040,
        ),
        (
            [
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.POWER_DOWN_GROUND.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
            ],
            0x0010,
        ),
        (
            [
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.POWER_DOWN_GROUND.value,
                PowerMode.NORMAL.value,
            ],
            0x0004,
        ),
        (
            [
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.NORMAL.value,
                PowerMode.POWER_DOWN_GROUND.value,
            ],
            0x0001,
        ),
    ],
)
def test_dac_generate_power_code(dac_state, expected, dac_ops):
    assert dac_ops.generate_power_code(dac_state) == expected


def test_dac_voltage_precisions(dac_ops):
    """checking if any mV value between 0-5V raises conversion error"""
    i = 0
    step = 0.001
    while i < 5.0:
        dac_ops.voltage_to_code(1, i, 1)
        i += step
    while i < 10.0:
        dac_ops.voltage_to_code(1, i, 2)
        i += step



"""unit tests for tc_commands.py module"""


from collections import Counter
from contextlib import nullcontext as does_not_raise

import pytest
from bitstring import BitArray
from edgepi.reg_helper.reg_helper import OpCode
from edgepi.tc.tc_constants import DecBits4, DecBits6, Masks, TCAddresses, TCType
from edgepi.tc.tc_commands import (
    ColdJunctionOverWriteError,
    IncompatibleRegisterSizeError,
    IncompleteTempError,
    MissingTCTypeError,
    TempOutOfRangeError,
    TempType,
    _dec_bits_to_float,
    code_to_temp,
    _negative_temp_check,
    tempcode_to_opcode,
    TempCode,
    _validate_temperatures,
)


@pytest.mark.parametrize(
    "code_bytes, temps",
    [
        ([0x0D, 0x88, 0x00, 0xC0, 0x00, 0x00], (-8, -1024)),  # negative temps
        ([0x0D, 0x08, 0x00, 0x40, 0x00, 0x00], (8, 1024)),  # positive temps
        (
            [0x0D, 0x7F, 0xFC, 0x7F, 0xFF, 0xE0],
            (127.984375, 2047.9921875),
        ),  # max temp values
        (
            [0x0D, 0xFF, 0xFC, 0xFF, 0xFF, 0xE0],
            (-127.984375, -2047.9921875),
        ),  # min temp values
        ([0x0D, 0x00, 0x00, 0x00, 0x00, 0x00], (0, 0)),  # zero temp
        ([0x0D, 0x15, 0x00, 0x01, 0x50, 0x00], (21, 21)),  # room temperature values
    ],
)
def test_code_to_temp(code_bytes, temps):
    assert code_to_temp(code_bytes) == temps


@pytest.mark.parametrize(
    "code_bytes, err_type",
    [
        ([0x0D], IndexError),  # should raise IndexError
        (
            [0x0D, "hello", "world", "!", 0x0F, 0x0F],
            ValueError,
        ),  # should raise ValueError
    ],
)
def test_code_to_temp_exceptions(code_bytes, err_type):
    with pytest.raises(Exception) as err:
        code_to_temp(code_bytes)
    assert err.type == err_type


@pytest.mark.parametrize(
    "temp_code, out, new_value",
    [
        (BitArray(uint=0b10001000, length=8), True, 0b1000),
        (BitArray(uint=0b01001000, length=8), False, 0b01001000),
    ],
)
def test_negative_temp_check(temp_code, out, new_value):
    assert _negative_temp_check(temp_code) == out
    assert temp_code.uint == new_value


@pytest.mark.parametrize(
    "tempcode, tc_type, opcode_list",
    [
        (
            TempCode(125, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            [OpCode(0x7D, TCAddresses.CJHF_W.value, Masks.BYTE_MASK.value)],
        ),  # single register, + value, no dec, no fillers
        (
            TempCode(-55, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            [OpCode(0xB7, TCAddresses.CJHF_W.value, Masks.BYTE_MASK.value)],
        ),  # single register, - value, no dec, no fillers
        (
            TempCode(
                0x4, DecBits4.P0_5, 3, 4, 0, TCAddresses.CJTO_W.value, TempType.COLD_JUNCTION_OFFSET
            ),
            TCType.TYPE_K,
            [OpCode(0x48, TCAddresses.CJTO_W.value, Masks.BYTE_MASK.value)],
        ),  # single register, + value, dec, no fillers
        (
            TempCode(
                -0x4,
                DecBits4.P0_5,
                3,
                4,
                0,
                TCAddresses.CJTO_W.value,
                TempType.COLD_JUNCTION_OFFSET,
            ),
            TCType.TYPE_K,
            [OpCode(0xC8, TCAddresses.CJTO_W.value, Masks.BYTE_MASK.value)],
        ),  # single register, - value, dec, no fillers
        (
            TempCode(
                1024, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            [
                OpCode(0x40, TCAddresses.LTHFTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x00, TCAddresses.LTHFTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, - value, no dec, no fillers
        (
            TempCode(-50, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE),
            TCType.TYPE_K,
            [
                OpCode(0x83, TCAddresses.LTHFTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x20, TCAddresses.LTHFTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, - value, no dec, no fillers
        (
            TempCode(
                1024, DecBits4.P0_75, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            [
                OpCode(0x40, TCAddresses.LTHFTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x0C, TCAddresses.LTHFTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, + value, dec, no fillers
        (
            TempCode(
                -50, DecBits4.P0_75, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            [
                OpCode(0x83, TCAddresses.LTHFTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x2C, TCAddresses.LTHFTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, - value, dec, no fillers
        (
            TempCode(
                1371, DecBits4.P0_9375, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            [
                OpCode(0x55, TCAddresses.LTHFTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0xBF, TCAddresses.LTHFTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, max + value, dec, no fillers
        (
            TempCode(
                64, DecBits6.P0_296875, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION
            ),
            TCType.TYPE_K,
            [
                OpCode(0x40, TCAddresses.CJTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x4C, TCAddresses.CJTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, + value, dec, fillers
        (
            TempCode(
                -8, DecBits6.P0_296875, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION
            ),
            TCType.TYPE_K,
            [
                OpCode(0x88, TCAddresses.CJTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x4C, TCAddresses.CJTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, - value, dec, fillers
        (
            TempCode(
                124, DecBits6.P0_984375, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION
            ),
            TCType.TYPE_K,
            [
                OpCode(0x7C, TCAddresses.CJTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0xFC, TCAddresses.CJTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, max + value, dec, fillers
        (
            TempCode(
                1, DecBits6.P0_015625, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION
            ),
            TCType.TYPE_K,
            [
                OpCode(0x01, TCAddresses.CJTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x04, TCAddresses.CJTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, min + value, dec, fillers
        (
            TempCode(0, DecBits6.P0, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            [
                OpCode(0x00, TCAddresses.CJTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x00, TCAddresses.CJTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, zero value, dec, fillers
        (
            TempCode(
                -1, DecBits6.P0_015625, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION
            ),
            TCType.TYPE_K,
            [
                OpCode(0x81, TCAddresses.CJTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0x04, TCAddresses.CJTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, max - value, dec, fillers
        (
            TempCode(
                -25, DecBits6.P0_984375, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION
            ),
            TCType.TYPE_K,
            [
                OpCode(0x99, TCAddresses.CJTH_W.value, Masks.BYTE_MASK.value),
                OpCode(0xFC, TCAddresses.CJTL_W.value, Masks.BYTE_MASK.value),
            ],
        ),  # 2 registers, min - value, dec, fillers
        (
            TempCode(None, None, 7, 6, 2, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            [],
        ),
    ],
)
def test_tempcode_to_opcode(tempcode, tc_type, opcode_list):
    result = tempcode_to_opcode(tempcode, tc_type, True)
    assert Counter(result) == Counter(opcode_list)


@pytest.mark.parametrize(
    "tempcode, tc_type, cj_status, err_type",
    [
        (
            TempCode(127, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            True,
            TempOutOfRangeError,
        ),
        (
            TempCode(127, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            True,
            TempOutOfRangeError,
        ),
        (None, TCType.TYPE_K, True, ValueError),
        (
            TempCode(None, DecBits6.P0, 7, 6, 2, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            True,
            IncompleteTempError,
        ),
        (
            TempCode(1000, None, 7, 6, 2, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            True,
            IncompleteTempError,
        ),
        (
            TempCode(120, DecBits4.P0, 8, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            True,
            IncompatibleRegisterSizeError,
        ),
        (
            TempCode(120, DecBits4.P0, 7, 1, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            True,
            IncompatibleRegisterSizeError,
        ),
        (
            TempCode(120, DecBits4.P0, 6, 1, 1, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            True,
            IncompatibleRegisterSizeError,
        ),
        (
            TempCode(127, DecBits4.P0, 8, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            None,
            True,
            MissingTCTypeError,
        ),
        (
            TempCode(120, DecBits4.P0, 6, 1, 1, TCAddresses.CJTH_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            False,
            ColdJunctionOverWriteError,
        ),
    ],
)
def test_tempcode_to_opcode_raises(tempcode, tc_type, cj_status, err_type):
    with pytest.raises(Exception) as err:
        tempcode_to_opcode(tempcode, tc_type, cj_status)
    assert err.type == err_type


@pytest.mark.parametrize(
    "tempcode, tc_type, expected",
    [
        (
            TempCode(126, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(125, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            does_not_raise(),
        ),
        (
            TempCode(-56, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(-55, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_K,
            does_not_raise(),
        ),
        (
            TempCode(
                1373, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(
                1372, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            does_not_raise(),
        ),
        (
            TempCode(
                -201, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(
                -200, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            does_not_raise(),
        ),
        (
            TempCode(126, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_B,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(125, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_B,
            does_not_raise(),
        ),
        (
            TempCode(-1, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_B,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(0, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_B,
            does_not_raise(),
        ),
        (
            TempCode(
                1821, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_B,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(
                1820, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_B,
            does_not_raise(),
        ),
        (
            TempCode(249, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE),
            TCType.TYPE_B,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(250, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE),
            TCType.TYPE_B,
            does_not_raise(),
        ),
        (
            TempCode(126, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_E,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(125, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_E,
            does_not_raise(),
        ),
        (
            TempCode(-56, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_E,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(-55, DecBits4.P0, 7, 0, 0, TCAddresses.CJHF_W.value, TempType.COLD_JUNCTION),
            TCType.TYPE_E,
            does_not_raise(),
        ),
        (
            TempCode(
                1001, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_E,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(
                1000, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_E,
            does_not_raise(),
        ),
        (
            TempCode(
                -201, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_E,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(
                -200, DecBits4.P0, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_E,
            does_not_raise(),
        ),
        (
            TempCode(
                1372, DecBits4.P0_25, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(
                1371, DecBits4.P0_4375, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            does_not_raise(),
        ),
        (
            TempCode(
                -200, DecBits4.P0_25, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(
                -199, DecBits4.P0_125, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_K,
            does_not_raise(),
        ),
        (
            TempCode(
                1820, DecBits4.P0_375, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_B,
            pytest.raises(TempOutOfRangeError),
        ),
        (
            TempCode(
                1819, DecBits4.P0_125, 11, 4, 0, TCAddresses.LTHFTH_W.value, TempType.THERMOCOUPLE
            ),
            TCType.TYPE_B,
            does_not_raise(),
        ),
    ],
)
def test_validate_temperatures(tempcode, tc_type, expected):
    with expected:
        _validate_temperatures(tempcode, tc_type)


@pytest.mark.parametrize(
    "dec_bits, num_dec_bits, float_val",
    [
        (DecBits4.P0.value, 4, 0),
        (DecBits4.P0_5.value, 4, 0.5),
        (DecBits4.P0_75.value, 4, 0.75),
        (DecBits4.P0_875.value, 4, 0.875),
        (DecBits4.P0_9375.value, 4, 0.9375),
        (DecBits4.P0_4375.value, 4, 0.4375),
        (DecBits4.P0_1875.value, 4, 0.1875),
        (DecBits4.P0_0625.value, 4, 0.0625),
        (DecBits4.P0_5625.value, 4, 0.5625),
        (DecBits4.P0_8125.value, 4, 0.8125),
        (DecBits4.P0_6875.value, 4, 0.6875),
        (DecBits4.P0_25.value, 4, 0.25),
        (DecBits4.P0_125.value, 4, 0.125),
        (DecBits4.P0_625.value, 4, 0.625),
        (DecBits4.P0_3125.value, 4, 0.3125),
        (DecBits4.P0_375.value, 4, 0.375),
        (DecBits6.P0.value, 6, 0),
        (DecBits6.P0_015625.value, 6, 0.015625),
        (DecBits6.P0_03125.value, 6, 0.03125),
        (DecBits6.P0_046875.value, 6, 0.046875),
        (DecBits6.P0_0625.value, 6, 0.0625),
        (DecBits6.P0_078125.value, 6, 0.078125),
        (DecBits6.P0_09375.value, 6, 0.09375),
        (DecBits6.P0_109375.value, 6, 0.109375),
        (DecBits6.P0_125.value, 6, 0.125),
        (DecBits6.P0_140625.value, 6, 0.140625),
        (DecBits6.P0_15625.value, 6, 0.15625),
        (DecBits6.P0_171875.value, 6, 0.171875),
        (DecBits6.P0_1875.value, 6, 0.1875),
        (DecBits6.P0_203125.value, 6, 0.203125),
        (DecBits6.P0_21875.value, 6, 0.21875),
        (DecBits6.P0_234375.value, 6, 0.234375),
        (DecBits6.P0_25.value, 6, 0.25),
        (DecBits6.P0_265625.value, 6, 0.265625),
        (DecBits6.P0_28125.value, 6, 0.28125),
        (DecBits6.P0_296875.value, 6, 0.296875),
        (DecBits6.P0_3125.value, 6, 0.3125),
        (DecBits6.P0_328125.value, 6, 0.328125),
        (DecBits6.P0_34375.value, 6, 0.34375),
        (DecBits6.P0_359375.value, 6, 0.359375),
        (DecBits6.P0_375.value, 6, 0.375),
        (DecBits6.P0_390625.value, 6, 0.390625),
        (DecBits6.P0_40625.value, 6, 0.40625),
        (DecBits6.P0_421875.value, 6, 0.421875),
        (DecBits6.P0_4375.value, 6, 0.4375),
        (DecBits6.P0_453125.value, 6, 0.453125),
        (DecBits6.P0_46875.value, 6, 0.46875),
        (DecBits6.P0_484375.value, 6, 0.484375),
        (DecBits6.P0_5.value, 6, 0.5),
        (DecBits6.P0_515625.value, 6, 0.515625),
        (DecBits6.P0_53125.value, 6, 0.53125),
        (DecBits6.P0_546875.value, 6, 0.546875),
        (DecBits6.P0_5625.value, 6, 0.5625),
        (DecBits6.P0_578125.value, 6, 0.578125),
        (DecBits6.P0_59375.value, 6, 0.59375),
        (DecBits6.P0_609375.value, 6, 0.609375),
        (DecBits6.P0_625.value, 6, 0.625),
        (DecBits6.P0_640625.value, 6, 0.640625),
        (DecBits6.P0_65625.value, 6, 0.65625),
        (DecBits6.P0_671875.value, 6, 0.671875),
        (DecBits6.P0_6875.value, 6, 0.6875),
        (DecBits6.P0_703125.value, 6, 0.703125),
        (DecBits6.P0_71875.value, 6, 0.71875),
        (DecBits6.P0_734375.value, 6, 0.734375),
        (DecBits6.P0_75.value, 6, 0.75),
        (DecBits6.P0_765625.value, 6, 0.765625),
        (DecBits6.P0_78125.value, 6, 0.78125),
        (DecBits6.P0_796875.value, 6, 0.796875),
        (DecBits6.P0_8125.value, 6, 0.8125),
        (DecBits6.P0_828125.value, 6, 0.828125),
        (DecBits6.P0_84375.value, 6, 0.84375),
        (DecBits6.P0_859375.value, 6, 0.859375),
        (DecBits6.P0_875.value, 6, 0.875),
        (DecBits6.P0_890625.value, 6, 0.890625),
        (DecBits6.P0_90625.value, 6, 0.90625),
        (DecBits6.P0_921875.value, 6, 0.921875),
        (DecBits6.P0_9375.value, 6, 0.9375),
        (DecBits6.P0_953125.value, 6, 0.953125),
        (DecBits6.P0_96875.value, 6, 0.96875),
        (DecBits6.P0_984375.value, 6, 0.984375),
    ],
)
def test_dec_bits_to_float(dec_bits, num_dec_bits, float_val):
    assert _dec_bits_to_float(dec_bits, num_dec_bits) == float_val



"""unit tests for adc_commands module"""


import pytest
from edgepi.adc.adc_commands import ADCCommands


@pytest.fixture(name="adc_ops")
def fixture_test_adc_ops():
    adc_ops = ADCCommands()
    return adc_ops


@pytest.mark.parametrize(
    "sample, result",
    [
        ([1], True),
        ([1, 2, 3, 4, 5], True),
        ([-1, 1, 2, 3], True),
        ([444, 22, 3333, 5], True),
        ([-111, -2222], True),
    ],
)
def test_check_for_int(sample, result, adc_ops):
    assert adc_ops.check_for_int(sample) == result


@pytest.mark.parametrize(
    "sample, error",
    [
        ([1, 2.22, 3, 4, 0], ValueError),
        ([None, 1, 2, 3], ValueError),
        ([], ValueError),
        ([-1, -2.22], ValueError),
    ],
)
def test_check_for_int_exception(sample, error, adc_ops):
    with pytest.raises(Exception) as err:
        adc_ops.check_for_int(sample)
    assert err.type is error


# Test read register content
# In: Register address and number of register to read
# Out: List of bytes Op-code and dummy bytes to transfer


@pytest.mark.parametrize(
    "address, num_of_regs, result",
    [
        (1, 1, [0x21, 0, 255]),
        (2, 4, [0x22, 3, 255, 255, 255, 255]),
        (3, 5, [0x23, 4, 255, 255, 255, 255, 255]),
        (10, 7, [0x2A, 6, 255, 255, 255, 255, 255, 255, 255]),
        (16, 10, [0x30, 9, 255, 255, 255, 255, 255, 255, 255, 255, 255, 255]),
        (16, 1, [0x30, 0, 255]),
    ],
)
def test_read_register_command(address, num_of_regs, result, adc_ops):
    assert adc_ops.read_register_command(address, num_of_regs) == result


@pytest.mark.parametrize(
    "address, num_of_regs, error", [(1, 3.2, ValueError), (3.2, 1, ValueError)]
)
def test_read_register_command_exception(address, num_of_regs, error, adc_ops):
    with pytest.raises(Exception) as err:
        adc_ops.read_register_command(address, num_of_regs)
        assert err.type is error


# Test write register content
# In: Register address and list of values to write
# Out: List of bytes Op-code and dummy bytes to transfer


@pytest.mark.parametrize(
    "address, values, result",
    [
        (1, [2], [0x41, 0, 2]),
        (2, [23, 25, 24, 102], [0x42, 3, 23, 25, 24, 102]),
        (3, [243, 123, 125, 111, 231], [0x43, 4, 243, 123, 125, 111, 231]),
        (10, [15, 25, 55, 225, 115, 2, 2], [0x4A, 6, 15, 25, 55, 225, 115, 2, 2]),
        (
            16,
            [25, 55, 5, 2, 5, 25, 55, 55, 25, 2],
            [0x50, 9, 25, 55, 5, 2, 5, 25, 55, 55, 25, 2],
        ),
    ],
)
def test_write_register_command(address, values, result, adc_ops):
    assert adc_ops.write_register_command(address, values) == result


@pytest.mark.parametrize(
    "address, values, error",
    [(1, [-1, 2], ValueError), (2, [23, 25, 256, 102], ValueError)],
)
def test_write_register_command_exception(address, values, error, adc_ops):
    with pytest.raises(Exception) as err:
        adc_ops.write_register_command(address, values)
        assert err.type is error



""" Constants for ADC device modules """


from dataclasses import dataclass
from enum import Enum, unique

from edgepi.reg_helper.reg_helper import BitMask, OpCode


ADC_NUM_REGS = 27  # number of ADC1263 registers
ADC_VOLTAGE_READ_LEN = 6  # number of bytes per voltage read


@unique
class ADCComs(Enum):
    """EdgePi ADC operation codes"""

    COM_NOP = 0x00
    COM_RESET = 0x06
    COM_START1 = 0x09
    COM_STOP1 = 0x0A
    COM_START2 = 0x0C
    COM_STOP2 = 0x0E
    COM_RDATA1 = 0x12
    COM_RDATA2 = 0x14
    COM_SYOCAL1 = 0x16
    COM_SYGCAL1 = 0x17
    COM_SFOCAL1 = 0x19
    COM_SYOCAL2 = 0x1B
    COM_SYGCAL2 = 0x1C
    COM_SFOCAL2 = 0x1E
    COM_RREG = 0x20
    COM_WREG = 0x40


@unique
class ADCReg(Enum):
    """EdgePi ADC register addresses"""

    REG_ID = 0x00
    REG_POWER = 0x01
    REG_INTERFACE = 0x02
    REG_MODE0 = 0x03
    REG_MODE1 = 0x04
    REG_MODE2 = 0x05
    REG_INPMUX = 0x06
    REG_OFCAL0 = 0x07
    REG_OFCAL1 = 0x08
    REG_OFCAL2 = 0x09
    REG_FSCAL0 = 0x0A
    REG_FSCAL1 = 0x0B
    REG_FSCAL2 = 0x0C
    REG_IDACMUX = 0x0D
    REG_IDACMAG = 0x0E
    REG_REFMUX = 0x0F
    REG_TDACP = 0x10
    REG_TDACN = 0x11
    REG_GPIOCON = 0x12
    REG_GPIODIR = 0x13
    REG_GPIODAT = 0x14
    REG_ADC2CFG = 0x15
    REG_ADC2MUX = 0x16
    REG_ADC2OFC0 = 0x17
    REG_ADC2OFC1 = 0x18
    REG_ADC2FSC0 = 0x19
    REG_ADC2FSC1 = 0x1A


@unique
class ADCChannel(Enum):
    """EdgePi ADC channels/channel modes"""

    AIN0 = 0
    AIN1 = 1
    AIN2 = 2
    AIN3 = 3
    AIN4 = 4
    AIN5 = 5
    AIN6 = 6
    AIN7 = 7
    AIN8 = 8
    AIN9 = 9
    AINCOM = 10
    FLOAT = 0xF


@dataclass(frozen=True)
class ADCReadInfo:
    """
    Data for use in performing ADC voltage reads

    Attributes:
        `id_num` (int): the ADS1263 ID number of this ADC
        `addx` (ADCReg): addx of this ADC's mux register
        `num_data_bytes` (int): number of data bytes for this ADC's reads
        `start_cmd` (ADCComs): command to trigger conversion(s) for this ADC
        `read_cmd` (ADCComs): command to read this ADC's data-holding register
        `stop_cmd` (ADCComs): command to stop conversion(s) for this ADC
    """

    id_num: int
    addx: ADCReg
    num_data_bytes: int
    start_cmd: ADCComs
    read_cmd: ADCComs
    stop_cmd: ADCComs


# TODO: use EEPROM values in ADCVoltageConfig
# TODO: refactor voltage configs from per ADC to per channel
@unique
class ADCNum(Enum):
    """ADS1263 ADCs"""

    ADC_1 = ADCReadInfo(
        1,
        ADCReg.REG_INPMUX,
        4,
        ADCComs.COM_START1.value,
        ADCComs.COM_RDATA1.value,
        ADCComs.COM_STOP1.value,
    )
    ADC_2 = ADCReadInfo(
        2,
        ADCReg.REG_ADC2MUX,
        3,
        ADCComs.COM_START2.value,
        ADCComs.COM_RDATA2.value,
        ADCComs.COM_STOP2.value,
    )


class ConvMode(Enum):
    """
    ADS1263 conversion modes for ADC1

    Attributes:

        `PULSE`: ADC1 performs a single conversion upon command

        `CONTINUOUS`: ADC1 peforms conversions continuously
    """

    CONTINUOUS = OpCode(0x00, ADCReg.REG_MODE0.value, BitMask.BIT6.value)
    PULSE = OpCode(0x40, ADCReg.REG_MODE0.value, BitMask.BIT6.value)


class ADC1DataRate(Enum):
    """ADS1263 data rates for ADC1"""

    SPS_2P5 = OpCode(0x0, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_5 = OpCode(0x1, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_10 = OpCode(0x2, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_16P6 = OpCode(0x3, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_20 = OpCode(0x4, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_50 = OpCode(0x5, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_60 = OpCode(0x6, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_100 = OpCode(0x7, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_400 = OpCode(0x8, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_1200 = OpCode(0x9, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_2400 = OpCode(0xA, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_4800 = OpCode(0xB, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_7200 = OpCode(0xC, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_14400 = OpCode(0xD, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_19200 = OpCode(0xE, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)
    SPS_38400 = OpCode(0xF, ADCReg.REG_MODE2.value, BitMask.LOW_NIBBLE.value)


class ADCMasks(Enum):
    """ADS1263 OpCode bit masks"""

    FILTER_BITS = 0x1F  # overwrite bits 7:5 (0-indexed)
    CHECK_BITS = 0xFC  # overwrite bits 1:0
    ADC2_DR_BITS = 0x3F  # overwrite bits 7:6
    RMUXP_BITS = 0xC7  # overwrite bits 5:3
    RMUXN_BITS = 0xF8  # overwrite bits 2:0


class ADC2DataRate(Enum):
    """ADS1263 data rates for ADC2"""

    SPS_10 = OpCode(0x00, ADCReg.REG_ADC2CFG.value, ADCMasks.ADC2_DR_BITS.value)
    SPS_100 = OpCode(0x40, ADCReg.REG_ADC2CFG.value, ADCMasks.ADC2_DR_BITS.value)
    SPS_400 = OpCode(0x80, ADCReg.REG_ADC2CFG.value, ADCMasks.ADC2_DR_BITS.value)
    SPS_800 = OpCode(0xC0, ADCReg.REG_ADC2CFG.value, ADCMasks.ADC2_DR_BITS.value)


class FilterMode(Enum):
    """ADC filter modes, for both ADC1 and ADC2"""

    SINC1 = OpCode(0x0, ADCReg.REG_MODE1.value, ADCMasks.FILTER_BITS.value)
    SINC2 = OpCode(0x20, ADCReg.REG_MODE1.value, ADCMasks.FILTER_BITS.value)
    SINC3 = OpCode(0x40, ADCReg.REG_MODE1.value, ADCMasks.FILTER_BITS.value)
    SINC4 = OpCode(0x60, ADCReg.REG_MODE1.value, ADCMasks.FILTER_BITS.value)
    FIR = OpCode(0x80, ADCReg.REG_MODE1.value, ADCMasks.FILTER_BITS.value)


class ADCReferenceSwitching(Enum):
    """Each ADC Reference ground can be configured"""

    GND_SW_NONE = 0
    GND_SW1 = 1
    GND_SW2 = 2
    GND_SW_BOTH = 3


class ADCPower(Enum):
    """OpCodes for configuring the ADS1263 POWER register"""

    RESET_CLEAR = OpCode(0x0, ADCReg.REG_POWER.value, BitMask.BIT4.value)


class CheckMode(Enum):
    """OpCodes for configuring ADS1263 voltage read check mode"""

    CHECK_BYTE_OFF = OpCode(0x00, ADCReg.REG_INTERFACE.value, ADCMasks.CHECK_BITS.value)
    CHECK_BYTE_CHK = OpCode(0x01, ADCReg.REG_INTERFACE.value, ADCMasks.CHECK_BITS.value)
    CHECK_BYTE_CRC = OpCode(0x02, ADCReg.REG_INTERFACE.value, ADCMasks.CHECK_BITS.value)


class StatusByte(Enum):
    """OpCodes for configuring ADS1263 voltage read status byte"""

    STATUS_BYTE_ON = OpCode(0x04, ADCReg.REG_INTERFACE.value, BitMask.BIT2.value)
    STATUS_BYTE_OFF = OpCode(0x00, ADCReg.REG_INTERFACE.value, BitMask.BIT2.value)


@dataclass(frozen=True)
class DifferentialPair:
    """ADC differential voltage reading channel pairs"""

    mux_p: ADCChannel
    mux_n: ADCChannel


class DiffMode(Enum):
    """ADC differential voltage reading modes"""

    DIFF_1 = DifferentialPair(ADCChannel.AIN0, ADCChannel.AIN1)
    DIFF_2 = DifferentialPair(ADCChannel.AIN2, ADCChannel.AIN3)
    DIFF_3 = DifferentialPair(ADCChannel.AIN4, ADCChannel.AIN5)
    DIFF_4 = DifferentialPair(ADCChannel.AIN6, ADCChannel.AIN7)
    DIFF_OFF = DifferentialPair(ADCChannel.FLOAT, ADCChannel.AINCOM)


class IDACMUX(Enum):
    """Settings for ADC IDACMUX register (output multiplexer mapping)"""

    IDAC2_AIN0 = OpCode(0x0, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN1 = OpCode(0x10, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN2 = OpCode(0x20, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN3 = OpCode(0x30, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN4 = OpCode(0x40, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN5 = OpCode(0x50, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN6 = OpCode(0x60, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN7 = OpCode(0x70, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN8 = OpCode(0x80, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AIN9 = OpCode(0x90, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_AINCOM = OpCode(0xA0, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_NO_CONNECT = OpCode(0xB0, ADCReg.REG_IDACMUX.value, BitMask.HIGH_NIBBLE.value)
    IDAC1_AIN0 = OpCode(0x0, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN1 = OpCode(0x1, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN2 = OpCode(0x2, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN3 = OpCode(0x3, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN4 = OpCode(0x4, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN5 = OpCode(0x5, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN6 = OpCode(0x6, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN7 = OpCode(0x7, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN8 = OpCode(0x8, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AIN9 = OpCode(0x9, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_AINCOM = OpCode(0xA, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)
    IDAC1_NO_CONNECT = OpCode(0xB, ADCReg.REG_IDACMUX.value, BitMask.LOW_NIBBLE.value)


class IDACMAG(Enum):
    """Settings for ADC IDACMAG register (IDAC current magnitude in µA)"""

    IDAC2_OFF = OpCode(0x0, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_50 = OpCode(0x10, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_100 = OpCode(0x20, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_250 = OpCode(0x30, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_500 = OpCode(0x40, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_750 = OpCode(0x50, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_1000 = OpCode(0x60, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_1500 = OpCode(0x70, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_2000 = OpCode(0x80, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_2500 = OpCode(0x90, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC2_3000 = OpCode(0xA0, ADCReg.REG_IDACMAG.value, BitMask.HIGH_NIBBLE.value)
    IDAC1_OFF = OpCode(0x0, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_50 = OpCode(0x1, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_100 = OpCode(0x2, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_250 = OpCode(0x3, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_500 = OpCode(0x4, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_750 = OpCode(0x5, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_1000 = OpCode(0x6, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_1500 = OpCode(0x7, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_2000 = OpCode(0x8, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_2500 = OpCode(0x9, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)
    IDAC1_3000 = OpCode(0xA, ADCReg.REG_IDACMAG.value, BitMask.LOW_NIBBLE.value)


class REFMUX(Enum):
    """Settings for ADC REFMUX register (Reference multiplexer mapping)"""

    POS_REF_INT_2P5 = OpCode(0x0, ADCReg.REG_REFMUX.value, ADCMasks.RMUXP_BITS.value)
    POS_REF_EXT_AIN0 = OpCode(0x8, ADCReg.REG_REFMUX.value, ADCMasks.RMUXP_BITS.value)
    POS_REF_EXT_AIN2 = OpCode(0x10, ADCReg.REG_REFMUX.value, ADCMasks.RMUXP_BITS.value)
    POS_REF_EXT_AIN4 = OpCode(0x18, ADCReg.REG_REFMUX.value, ADCMasks.RMUXP_BITS.value)
    POS_REF_INT_VAVDD = OpCode(0x20, ADCReg.REG_REFMUX.value, ADCMasks.RMUXP_BITS.value)
    NEG_REF_INT_2P5 = OpCode(0x0, ADCReg.REG_REFMUX.value, ADCMasks.RMUXN_BITS.value)
    NEG_REF_EXT_AIN1 = OpCode(0x1, ADCReg.REG_REFMUX.value, ADCMasks.RMUXN_BITS.value)
    NEG_REF_EXT_AIN3 = OpCode(0x2, ADCReg.REG_REFMUX.value, ADCMasks.RMUXN_BITS.value)
    NEG_REF_EXT_AIN5 = OpCode(0x3, ADCReg.REG_REFMUX.value, ADCMasks.RMUXN_BITS.value)
    NEG_REF_INT_VAVSS = OpCode(0x4, ADCReg.REG_REFMUX.value, ADCMasks.RMUXN_BITS.value)


class RTDModes(Enum):
    "ADC.__config args for for turning RTD mode on/off"
    RTD_ON = {
        "adc_1_analog_in": ADCChannel.AIN5,
        "adc_1_mux_n": ADCChannel.AIN6,
        "idac_1_mux": IDACMUX.IDAC1_AIN8,
        "idac_2_mux": IDACMUX.IDAC2_AIN9,
        "idac_1_mag": IDACMAG.IDAC1_500,
        "idac_2_mag": IDACMAG.IDAC2_500,
        "pos_ref_inp": REFMUX.POS_REF_EXT_AIN4,
        "neg_ref_inp": REFMUX.NEG_REF_INT_VAVSS,
    }
    RTD_OFF = {
        "adc_1_analog_in": ADCChannel.FLOAT,
        "adc_1_mux_n": ADCChannel.AINCOM,
        "idac_1_mux": IDACMUX.IDAC1_NO_CONNECT,
        "idac_2_mux": IDACMUX.IDAC2_NO_CONNECT,
        "idac_1_mag": IDACMAG.IDAC1_OFF,
        "idac_2_mag": IDACMAG.IDAC2_OFF,
        "pos_ref_inp": REFMUX.POS_REF_INT_2P5,
        "neg_ref_inp": REFMUX.NEG_REF_INT_2P5,
    }


class AllowedChannels(Enum):
    """Available channels for reading depend on whether RTD is enabled or not"""

    RTD_ON = [
        ADCChannel.AIN0,
        ADCChannel.AIN1,
        ADCChannel.AIN2,
        ADCChannel.AIN3,
        ADCChannel.AIN7,
        ADCChannel.AINCOM,
        ADCChannel.FLOAT,
    ]
    RTD_OFF = [
        ADCChannel.AIN0,
        ADCChannel.AIN1,
        ADCChannel.AIN2,
        ADCChannel.AIN3,
        ADCChannel.AIN4,
        ADCChannel.AIN5,
        ADCChannel.AIN6,
        ADCChannel.AIN7,
        ADCChannel.AINCOM,
        ADCChannel.FLOAT,
    ]



""" Utility module containing constants for gpio modules """


from enum import Enum, unique
from edgepi.reg_helper.reg_helper import OpCode, BitMask


class GpioExpanderAddress(Enum):
    """GPIO expander i2c device addresses"""

    EXP_ONE = 32
    EXP_TWO = 33


@unique
class GPIOAddresses(Enum):
    """GPIO expander port addresses"""

    # Read addresses
    INPUT_PORT_0 = 0x00
    INPUT_PORT_1 = 0x01
    OUTPUT_PORT_0 = 0x02    # Port B
    OUTPUT_PORT_1 = 0x03    # Port A
    POLARITY_INVERSION_PORT_0 = 0x04
    POLARITY_INVERSION_PORT_1 = 0x05
    CONFIGURATION_PORT_0 = 0x06 # Port B
    CONFIGURATION_PORT_1 = 0x07 # Port A


@unique
class GpioAOutputSet(Enum):
    """valid opcodes for setting a GPIO Expander's Port A (Output Port 1)"""

    SET_OUTPUT_1 = OpCode(0x01, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT0.value)
    SET_OUTPUT_2 = OpCode(0x02, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT1.value)
    SET_OUTPUT_3 = OpCode(0x04, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT2.value)
    SET_OUTPUT_4 = OpCode(0x08, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT3.value)
    SET_OUTPUT_5 = OpCode(0x10, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT4.value)
    SET_OUTPUT_6 = OpCode(0x20, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT5.value)
    SET_OUTPUT_7 = OpCode(0x40, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT6.value)
    SET_OUTPUT_8 = OpCode(0x80, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT7.value)
    SET_OUTPUT_ALL = OpCode(0xFF, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BYTE.value)


@unique
class GpioAOutputClear(Enum):
    """valid opcodes for clearing a GPIO Expander's A (Output Port 1)"""

    CLEAR_OUTPUT_1 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT0.value)
    CLEAR_OUTPUT_2 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT1.value)
    CLEAR_OUTPUT_3 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT2.value)
    CLEAR_OUTPUT_4 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT3.value)
    CLEAR_OUTPUT_5 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT4.value)
    CLEAR_OUTPUT_6 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT5.value)
    CLEAR_OUTPUT_7 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT6.value)
    CLEAR_OUTPUT_8 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_1.value, BitMask.BIT7.value)
    CLEAR_OUTPUT_ALL = OpCode(
        0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BYTE.value
    )


@unique
class GpioBOutputSet(Enum):
    """valid opcodes for setting a GPIO Expander's Port B (Output Port 0)"""

    SET_OUTPUT_1 = OpCode(0x01, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT0.value)
    SET_OUTPUT_2 = OpCode(0x02, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT1.value)
    SET_OUTPUT_3 = OpCode(0x04, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT2.value)
    SET_OUTPUT_4 = OpCode(0x08, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT3.value)
    SET_OUTPUT_5 = OpCode(0x10, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT4.value)
    SET_OUTPUT_6 = OpCode(0x20, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT5.value)
    SET_OUTPUT_7 = OpCode(0x40, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT6.value)
    SET_OUTPUT_8 = OpCode(0x80, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT7.value)
    SET_OUTPUT_ALL = OpCode(0xFF, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BYTE.value)


@unique
class GpioBOutputClear(Enum):
    """valid opcodes for clearing a GPIO Expander's Port B (Output Port 0)"""

    CLEAR_OUTPUT_1 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT0.value)
    CLEAR_OUTPUT_2 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT1.value)
    CLEAR_OUTPUT_3 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT2.value)
    CLEAR_OUTPUT_4 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT3.value)
    CLEAR_OUTPUT_5 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT4.value)
    CLEAR_OUTPUT_6 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT5.value)
    CLEAR_OUTPUT_7 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT6.value)
    CLEAR_OUTPUT_8 = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BIT7.value)
    CLEAR_OUTPUT_ALL = OpCode(0x00, GPIOAddresses.OUTPUT_PORT_0.value, BitMask.BYTE.value)

@unique
class GpioAPinDirOut(Enum):
    """
    Opcodes for setting a GPIO Expander's Port A pin direction to output (Configuration Port 1)
    """
    PIN1_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT0.value)
    PIN2_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT1.value)
    PIN3_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT2.value)
    PIN4_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT3.value)
    PIN5_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT4.value)
    PIN6_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT5.value)
    PIN7_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT6.value)
    PIN8_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT7.value)
    ALL_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BYTE.value)

@unique
class GpioAPinDirIn(Enum):
    """
    Opcodes for setting a GPIO Expander's Port A pin direction to input (Configuration Port 1)
    """
    PIN1_DIR_IN = OpCode(0x01, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT0.value)
    PIN2_DIR_IN = OpCode(0x02, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT1.value)
    PIN3_DIR_IN = OpCode(0x04, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT2.value)
    PIN4_DIR_IN = OpCode(0x08, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT3.value)
    PIN5_DIR_IN = OpCode(0x10, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT4.value)
    PIN6_DIR_IN = OpCode(0x20, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT5.value)
    PIN7_DIR_IN = OpCode(0x40, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT6.value)
    PIN8_DIR_IN = OpCode(0x80, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BIT7.value)
    ALL_DIR_IN = OpCode(0xFF, GPIOAddresses.CONFIGURATION_PORT_1.value, BitMask.BYTE.value)


@unique
class GpioBPinDirOut(Enum):
    """
    Opcodes for setting a GPIO Expander's Port B pin direction to output (Configuration Port 0)
    """
    PIN1_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT0.value)
    PIN2_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT1.value)
    PIN3_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT2.value)
    PIN4_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT3.value)
    PIN5_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT4.value)
    PIN6_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT5.value)
    PIN7_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT6.value)
    PIN8_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT7.value)
    ALL_DIR_OUT = OpCode(0x00, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BYTE.value)

@unique
class GpioBPinDirIn(Enum):
    """
    Opcodes for setting a GPIO Expander's Port B pin direction to input (Configuration Port 0)
    """
    PIN1_DIR_IN = OpCode(0x01, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT0.value)
    PIN2_DIR_IN = OpCode(0x02, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT1.value)
    PIN3_DIR_IN = OpCode(0x04, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT2.value)
    PIN4_DIR_IN = OpCode(0x08, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT3.value)
    PIN5_DIR_IN = OpCode(0x10, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT4.value)
    PIN6_DIR_IN = OpCode(0x20, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT5.value)
    PIN7_DIR_IN = OpCode(0x40, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT6.value)
    PIN8_DIR_IN = OpCode(0x80, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BIT7.value)
    ALL_DIR_IN = OpCode(0xFF, GPIOAddresses.CONFIGURATION_PORT_0.value, BitMask.BYTE.value)

@unique
class GpioDevPaths(Enum):
    """
    GPIO device paths
    """
    I2C_DEV_PATH = '/dev/i2c-10'
    GPIO_CIHP_DEV_PATH = '/dev/gpiochip0'

@unique
class GpioPins(Enum):
    """GPIO Pin Names"""
    LED_OVR1 = 'LED_OVR1'
    LED_OVR2 = 'LED_OVR2'
    LED_OVR3 = 'LED_OVR3'
    LED_OVR4 = 'LED_OVR4'
    LED_OVR5 = 'LED_OVR5'
    LED_OVR6 = 'LED_OVR6'
    LED_OVR7 = 'LED_OVR7'
    LED_OVR8 = 'LED_OVR8'
    AO_EN1 = 'AO_EN1'
    AO_EN2 = 'AO_EN2'
    AO_EN3 = 'AO_EN3'
    AO_EN4 = 'AO_EN4'
    AO_EN5 = 'AO_EN5'
    AO_EN6 = 'AO_EN6'
    AO_EN7 = 'AO_EN7'
    AO_EN8 = 'AO_EN8'
    DAC_GAIN = 'DAC_GAIN'
    GNDSW_IN1 = 'GNDSW_IN1'
    GNDSW_IN2 = 'GNDSW_IN2'
    RTD_EN = 'RTD_EN'
    RELAY = 'RELAY'
    DOUT1 = 'DOUT1'
    DOUT2 = 'DOUT2'
    DOUT3 = 'DOUT3'
    DOUT4 = 'DOUT4'
    DOUT5 = 'DOUT5'
    DOUT6 = 'DOUT6'
    DOUT7 = 'DOUT7'
    DOUT8 = 'DOUT8'
    DIN1 = 'DIN1'
    DIN2 = 'DIN2'
    DIN3 = 'DIN3'
    DIN4 = 'DIN4'
    DIN5 = 'DIN5'
    DIN6 = 'DIN6'
    DIN7 = 'DIN7'
    DIN8 = 'DIN8'
    BUZZER = 'BUZZER'
    AUDIO = 'AUDIO'




"""Digital Output Module"""

from edgepi.gpio.gpio_constants import GpioPins
from edgepi.gpio.gpio_configs import DOUTPins
from edgepi.gpio.edgepi_gpio import EdgePiGPIO

class InvalidPinName(Exception):
    """Raised invalid pin name"""

class EdgePiDigitalOutput():
    """handling digital output"""
    def __init__(self):
        # To limit access to input functionality, using composition rather than inheritance
        self.gpio = EdgePiGPIO()

    def digital_output_state(self, pin_name: GpioPins = None, state: bool = None):
        """
        change the output state of the pin to the state passed as argument
        Args:
            pin_name (GpioPins): GpioPin enums
            state (bool): True = output high, False, output low
        """
        if state is None:
            raise ValueError(f'Invalid state passed: {state}')
        if pin_name is None or pin_name.value not in [pins.value for pins in DOUTPins]:
            raise InvalidPinName(f'Invalid pin name passed: {pin_name}')
        if state:
            self.gpio.set_pin_state(pin_name.value)
        else:
            self.gpio.clear_pin_state(pin_name.value)

    def digital_output_direction(self, pin_name: GpioPins = None, direction: bool = None):
        """
        change the output state of the pin to the state passed as argument
        Args:
            pin_name (GpioPins): GpioPin enums
            state (bool): True = direction input, False = direction output
        """
        if direction is None:
            raise ValueError(f'Invalid direction passed: {direction}')
        if pin_name is None or pin_name.value not in [pins.value for pins in DOUTPins]:
            raise InvalidPinName(f'Invalid pin name passed: {pin_name}')
        if direction:
            self.gpio.set_pin_direction_in(pin_name.value)
        else:
            self.gpio.set_pin_direction_out(pin_name.value)



"""unit tests for tc_conv_time module"""


import pytest
from edgepi.tc.tc_conv_time import SAFETY_MARGIN, calc_conv_time, ConvTimes

# pylint: disable=too-many-lines

@pytest.mark.parametrize(
    "cr0_val, cr1_val, safe_delay, expected",
    [
        ########################## 60 hz, max base time
        (0x00, 0x03, True, ConvTimes.DEFAULT_HZ60_MAX.value),  # AV_1, OC_OFF, CJ_ON, DEF_MAX, 60HZ
        (
            0x00,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.HZ60_AVG_2.value,
        ),  # AV_2, OC_OFF, CJ_ON, DEF_MAX, 60HZ
        (
            0x00,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.HZ60_AVG_4.value,
        ),  # AV_4, OC_OFF, CJ_ON, DEF_MAX, 60HZ
        (
            0x00,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.HZ60_AVG_8.value,
        ),  # AV_8, OC_OFF, CJ_ON, DEF_MAX, 60HZ
        (
            0x00,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.HZ60_AVG_16.value,
        ),  # AV_16, OC_OFF, CJ_ON, DEF_MAX, 60HZ
        (
            0x10,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_1, OC_L_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x10,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_2, OC_L_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x10,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_4, OC_L_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x10,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_8, OC_L_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x10,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_16, OC_L_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x20,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_1, OC_M_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x20,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_2, OC_M_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x20,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_4, OC_M_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x20,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_8, OC_M_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x20,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_16, OC_M_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x30,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_1, OC_H_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x30,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_2, OC_H_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x30,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_4, OC_H_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x30,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_8, OC_H_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x30,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_16, OC_H_MAX, CJ_ON, DEF_MAX, 60HZ
        (
            0x08,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_OFF, CJ_OFF, DEF_MAX, 60HZ
        (
            0x08,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.HZ60_AVG_2.value + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_OFF, CJ_OFF, DEF_MAX, 60HZ
        (
            0x08,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.HZ60_AVG_4.value + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_OFF, CJ_OFF, DEF_MAX, 60HZ
        (
            0x08,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.HZ60_AVG_8.value + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_OFF, CJ_OFF, DEF_MAX, 60HZ
        (
            0x08,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value + ConvTimes.HZ60_AVG_16.value + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_OFF, CJ_OFF, DEF_MAX, 60HZ
        (
            0x18,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_L_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x18,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_L_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x18,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_L_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x18,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_L_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x18,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_L_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x28,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_M_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x28,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_M_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x28,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_M_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x28,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_M_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x28,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_M_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x38,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_H_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x38,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_H_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x38,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_H_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x38,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_H_MAX, CJ_OFF, DEF_MAX, 60HZ
        (
            0x38,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ60_MAX.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_H_MAX, CJ_OFF, DEF_MAX, 60HZ
        ########################## 60 hz, nominal base time
        (0x00, 0x03, False, ConvTimes.DEFAULT_HZ60_NOM.value),  # AV_1, OC_OFF, CJ_ON, DEF_NOM, 60HZ
        (
            0x00,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.HZ60_AVG_2.value,
        ),  # AV_2, OC_OFF, CJ_ON, DEF_NOM, 60HZ
        (
            0x00,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.HZ60_AVG_4.value,
        ),  # AV_4, OC_OFF, CJ_ON, DEF_NOM, 60HZ
        (
            0x00,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.HZ60_AVG_8.value,
        ),  # AV_8, OC_OFF, CJ_ON, DEF_NOM, 60HZ
        (
            0x00,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.HZ60_AVG_16.value,
        ),  # AV_16, OC_OFF, CJ_ON, DEF_NOM, 60HZ
        (
            0x10,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_1, OC_L_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x10,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_2, OC_L_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x10,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_4, OC_L_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x10,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_8, OC_L_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x10,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_16, OC_L_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x20,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_1, OC_M_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x20,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_2, OC_M_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x20,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_4, OC_M_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x20,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_8, OC_M_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x20,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_16, OC_M_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x30,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_1, OC_H_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x30,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_2, OC_H_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x30,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_4, OC_H_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x30,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_8, OC_H_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x30,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_16, OC_H_NOM, CJ_ON, DEF_NOM, 60HZ
        (
            0x08,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_OFF, CJ_OFF, DEF_NOM, 60HZ
        (
            0x08,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.HZ60_AVG_2.value + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_OFF, CJ_OFF, DEF_NOM, 60HZ
        (
            0x08,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.HZ60_AVG_4.value + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_OFF, CJ_OFF, DEF_NOM, 60HZ
        (
            0x08,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.HZ60_AVG_8.value + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_OFF, CJ_OFF, DEF_NOM, 60HZ
        (
            0x08,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value + ConvTimes.HZ60_AVG_16.value + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_OFF, CJ_OFF, DEF_NOM, 60HZ
        (
            0x18,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_L_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x18,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_L_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x18,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_L_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x18,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_L_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x18,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_L_NOM, CJ_OFF, DEF_V, 60HZ
        (
            0x28,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_M_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x28,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_M_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x28,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_M_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x28,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_M_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x28,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_M_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x38,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_H_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x38,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_2.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_H_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x38,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_4.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_H_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x38,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_8.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_H_NOM, CJ_OFF, DEF_NOM, 60HZ
        (
            0x38,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ60_NOM.value
            + ConvTimes.HZ60_AVG_16.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_H_NOM, CJ_OFF, DEF_NOM, 60HZ
        ########################## 50 hz, max base time
        (0x01, 0x03, True, ConvTimes.DEFAULT_HZ50_MAX.value),  # AV_1, OC_OFF, CJ_ON, DEF_MAX, 50HZ
        (
            0x01,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.HZ50_AVG_2.value,
        ),  # AV_2, OC_OFF, CJ_ON, DEF_MAX, 50HZ
        (
            0x01,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.HZ50_AVG_4.value,
        ),  # AV_4, OC_OFF, CJ_ON, DEF_MAX, 50HZ
        (
            0x01,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.HZ50_AVG_8.value,
        ),  # AV_8, OC_OFF, CJ_ON, DEF_MAX, 50HZ
        (
            0x01,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.HZ50_AVG_16.value,
        ),  # AV_16, OC_OFF, CJ_ON, DEF_MAX, 50HZ
        (
            0x11,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_1, OC_L_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x11,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_2, OC_L_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x11,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_4, OC_L_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x11,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_8, OC_L_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x11,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_ON_OC_LOW_MAX.value,
        ),  # AV_16, OC_L_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x21,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_1, OC_M_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x21,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_2, OC_M_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x21,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_4, OC_M_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x21,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_8, OC_M_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x21,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_ON_OC_MED_MAX.value,
        ),  # AV_16, OC_M_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x31,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_1, OC_H_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x31,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_2, OC_H_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x31,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_4, OC_H_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x31,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_8, OC_H_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x31,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_ON_OC_HIGH_MAX.value,
        ),  # AV_16, OC_H_MAX, CJ_ON, DEF_MAX, 50HZ
        (
            0x09,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_OFF, CJ_OFF, DEF_MAX, 50HZ
        (
            0x09,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.HZ50_AVG_2.value + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_OFF, CJ_OFF, DEF_MAX, 50HZ
        (
            0x09,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.HZ50_AVG_4.value + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_OFF, CJ_OFF, DEF_MAX, 50HZ
        (
            0x09,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.HZ50_AVG_8.value + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_OFF, CJ_OFF, DEF_MAX, 50HZ
        (
            0x09,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value + ConvTimes.HZ50_AVG_16.value + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_OFF, CJ_OFF, DEF_MAX, 50HZ
        (
            0x19,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_L_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x19,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_L_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x19,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_L_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x19,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_L_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x19,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_OFF_OC_LOW_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_L_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x29,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_M_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x29,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_M_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x29,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_M_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x29,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_M_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x29,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_OFF_OC_MED_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_M_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x39,
            0x03,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_H_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x39,
            0x13,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_H_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x39,
            0x23,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_H_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x39,
            0x33,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_H_MAX, CJ_OFF, DEF_MAX, 50HZ
        (
            0x39,
            0x43,
            True,
            ConvTimes.DEFAULT_HZ50_MAX.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_OFF_OC_HIGH_MAX.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_H_MAX, CJ_OFF, DEF_MAX, 50HZ
        ########################## 50 hz, nominal base time
        (0x01, 0x03, False, ConvTimes.DEFAULT_HZ50_NOM.value),  # AV_1, OC_OFF, CJ_ON, DEF_NOM, 50HZ
        (
            0x01,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.HZ50_AVG_2.value,
        ),  # AV_2, OC_OFF, CJ_ON, DEF_NOM, 50HZ
        (
            0x01,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.HZ50_AVG_4.value,
        ),  # AV_4, OC_OFF, CJ_ON, DEF_NOM, 50HZ
        (
            0x01,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.HZ50_AVG_8.value,
        ),  # AV_8, OC_OFF, CJ_ON, DEF_NOM, 50HZ
        (
            0x01,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.HZ50_AVG_16.value,
        ),  # AV_16, OC_OFF, CJ_ON, DEF_NOM, 50HZ
        (
            0x11,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_1, OC_L_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x11,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_2, OC_L_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x11,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_4, OC_L_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x11,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_8, OC_L_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x11,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_ON_OC_LOW_NOM.value,
        ),  # AV_16, OC_L_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x21,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_1, OC_M_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x21,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_2, OC_M_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x21,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_4, OC_M_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x21,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_8, OC_M_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x21,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_ON_OC_MED_NOM.value,
        ),  # AV_16, OC_M_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x31,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_1, OC_H_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x31,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_2, OC_H_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x31,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_4, OC_H_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x31,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_8, OC_H_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x31,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_ON_OC_HIGH_NOM.value,
        ),  # AV_16, OC_H_NOM, CJ_ON, DEF_NOM, 50HZ
        (
            0x09,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_OFF, CJ_OFF, DEF_NOM, 50HZ
        (
            0x09,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.HZ50_AVG_2.value + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_OFF, CJ_OFF, DEF_NOM, 50HZ
        (
            0x09,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.HZ50_AVG_4.value + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_OFF, CJ_OFF, DEF_NOM, 50HZ
        (
            0x09,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.HZ50_AVG_8.value + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_OFF, CJ_OFF, DEF_NOM, 50HZ
        (
            0x09,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value + ConvTimes.HZ50_AVG_16.value + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_OFF, CJ_OFF, DEF_NOM, 50HZ
        (
            0x19,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_L_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x19,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_L_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x19,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_L_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x19,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_L_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x19,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_OFF_OC_LOW_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_L_NOM, CJ_OFF, DEF_V, 50HZ
        (
            0x29,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_M_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x29,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_M_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x29,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_M_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x29,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_M_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x29,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_OFF_OC_MED_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_M_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x39,
            0x03,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_1, OC_H_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x39,
            0x13,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_2.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_2, OC_H_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x39,
            0x23,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_4.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_4, OC_H_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x39,
            0x33,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_8.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_8, OC_H_NOM, CJ_OFF, DEF_NOM, 50HZ
        (
            0x39,
            0x43,
            False,
            ConvTimes.DEFAULT_HZ50_NOM.value
            + ConvTimes.HZ50_AVG_16.value
            + ConvTimes.CJ_OFF_OC_HIGH_NOM.value
            + ConvTimes.CJ_OFF.value,
        ),  # AV_16, OC_H_NOM, CJ_OFF, DEF_NOM, 50HZ
    ],
)
def test_calc_conv_time(cr0_val, cr1_val, safe_delay, expected):
    assert calc_conv_time(cr0_val, cr1_val, safe_delay) == pytest.approx(expected * SAFETY_MARGIN)



""" utility module for general purpose functions and classes

    Functions:
        filter_dict(dict, any)
        bitstring_from_list(list)
"""


from bitstring import BitArray, pack


def filter_dict(dictionary: dict, entry_key="", entry_val="") -> dict:
    """use for filtering an entry from a dictionary by key or value

    Args:
        dictionary (dict): any dictionary whose entries are to be filtered

        entry_key (any): the key of the entry to filter out

        entry_val (any): the value of the entry or entries to filter out

    Returns:
        a dictionary of entries from the original dictionary, after filtering out entries whose
        key or value matches either the entry_key or entry_val, respectively.
    """
    filtered_args = {
        key: value for (key, value) in dictionary.items() if key != entry_key and value != entry_val
    }
    return filtered_args


def bitstring_from_list(data: list) -> BitArray:
    """
    Builds a bitstring from a list of uint byte values

    Args:
        data (List[int]): a list of uint byte values

    Returns:
        BitArray: bitstring of bytes ordered from data[0], data[1], ..., data[n-1]
    """
    code = BitArray()
    for value in data:
        next_byte = pack("uint:8", value)
        code.append(next_byte)
    return code
